//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
/*
 * gtest_force_torque_inertial_tools.cpp
 *
 *  Created on: Aug 8, 2021
 *      Author: jsola
 */

#include "bodydynamics/math/force_torque_inertial_delta_tools.h"
#include <core/utils/utils_gtest.h>

#include <iostream>
#include <iomanip>

using namespace Eigen;
using namespace wolf;
using namespace bodydynamics;
using namespace fti;

TEST(FTI_tools, identity)
{
    VectorXd id_true;
    id_true.setZero(19);
    id_true(18) = 1.0;

    VectorXd id = identity<double>();
    ASSERT_MATRIX_APPROX(id, id_true, 1e-10);
}

TEST(FTI_tools, identity_split)
{
    VectorXd    pi(3), vi(3), pd(3), vd(3), L(3), qv(4);
    Quaterniond q;

    identity(pi, vi, pd, vd, L, q);
    ASSERT_MATRIX_APPROX(pi, Vector3d::Zero(), 1e-10);
    ASSERT_MATRIX_APPROX(vi, Vector3d::Zero(), 1e-10);
    ASSERT_MATRIX_APPROX(pd, Vector3d::Zero(), 1e-10);
    ASSERT_MATRIX_APPROX(vd, Vector3d::Zero(), 1e-10);
    ASSERT_MATRIX_APPROX(L, Vector3d::Zero(), 1e-10);
    ASSERT_QUATERNION_APPROX(q, Quaterniond::Identity(), 1e-10);

    identity(pi, vi, pd, vd, L, qv);
    ASSERT_MATRIX_APPROX(pi, Vector3d::Zero(), 1e-10);
    ASSERT_MATRIX_APPROX(vi, Vector3d::Zero(), 1e-10);
    ASSERT_MATRIX_APPROX(pd, Vector3d::Zero(), 1e-10);
    ASSERT_MATRIX_APPROX(vd, Vector3d::Zero(), 1e-10);
    ASSERT_MATRIX_APPROX(L, Vector3d::Zero(), 1e-10);
    ASSERT_MATRIX_APPROX(qv, (Vector4d() << 0, 0, 0, 1).finished(), 1e-10);
}

TEST(FTI_tools, inverse)
{
    VectorXd d(19), id(19), iid(19), iiid(19), I(19);
    Vector4d qv;
    double   dt = 0.1;

    qv = (Vector4d() << 9, 10, 11, 12).finished().normalized();
    d << 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, qv;

    id = inverse(d, dt);

    compose(id, d, dt, I);
    ASSERT_MATRIX_APPROX(I, identity(), 1e-10);
    compose(d, id, -dt, I);  // Observe -dt is reversed !!
    ASSERT_MATRIX_APPROX(I, identity(), 1e-10);

    inverse(id, -dt, iid);  // Observe -dt is reversed !!
    ASSERT_MATRIX_APPROX(d, iid, 1e-10);
    iiid = inverse(iid, dt);
    ASSERT_MATRIX_APPROX(id, iiid, 1e-10);
}

TEST(FTI_tools, compose_between)
{
    VectorXd              dx1(19), dx2(19), dx3(19);
    Matrix<double, 19, 1> d1, d2, d3;
    double                dt = 0.1;

    dx1 << VectorXd::Random(15), Vector4d::Random().normalized();
    d1 = dx1;
    dx2 << VectorXd::Random(15), Vector4d::Random().normalized();
    d2 = dx2;

    // combinations of templates for sum()
    compose(dx1, dx2, dt, dx3);
    compose(d1, d2, dt, d3);
    ASSERT_MATRIX_APPROX(d3, dx3, 1e-10);

    compose(d1, dx2, dt, dx3);
    d3 = compose(dx1, d2, dt);
    ASSERT_MATRIX_APPROX(d3, dx3, 1e-10);

    // // minus(d1, d3) should recover delta_2
    Matrix<double, 19, 1> bet;
    between(d1, d3, dt, bet);
    ASSERT_MATRIX_APPROX(bet, d2, 1e-10);

    // minus(d3, d1, -dt) should recover inverse(d2, dt)
    bet = between(d3, d1, -dt);
    ASSERT_MATRIX_APPROX(bet, inverse(d2, dt), 1e-10);
};

TEST(FTI_tools, compose_between_with_state)
{
    VectorXd x(13), d(19), x2(13), x3(13), d2(19), d3(19);
    double   dt = 0.1;

    x << VectorXd::Random(9), Vector4d::Random().normalized();
    d << VectorXd::Random(15), Vector4d::Random().normalized();

    {
        // 1. compose over state with IMU delta
        composeOverState(x, d, dt, x2);
        x3 = composeOverState(x, d, dt);
        ASSERT_MATRIX_APPROX(x3, x2, 1e-10);

        // betweenStates(x, x2) should recover d
        betweenStates(x, x2, dt, d2);
        d3 = betweenStates(x, x2, dt);
        ASSERT_MATRIX_APPROX(d2, d3, 1e-10);
        VectorComposite dc(d, "PVpvLO", {3, 3, 3, 3, 3, 4});
        VectorComposite d2c(d2, "PVpvLO", {3, 3, 3, 3, 3, 4});
        VectorComposite d3c(d3, "PVpvLO", {3, 3, 3, 3, 3, 4});
        VectorXd        de  = dc.vector("PVLO");
        VectorXd        d2e = d2c.vector("PVLO");
        VectorXd        d3e = d3c.vector("PVLO");
        ASSERT_MATRIX_APPROX(d2c.vector("PVLO"), dc.vector("PVLO"), 1e-10);
        ASSERT_MATRIX_APPROX(d3c.vector("PVLO"), dc.vector("PVLO"), 1e-10);
    }
    {
        // 2. compose over state with FT delta
        composeOverState(x, d, dt, x2, fti::FT);
        x3 = composeOverState(x, d, dt, fti::FT);
        ASSERT_MATRIX_APPROX(x3, x2, 1e-10);

        // betweenStates(x, x2) should recover d
        betweenStates(x, x2, dt, d2);
        d3 = betweenStates(x, x2, dt);
        ASSERT_MATRIX_APPROX(d2, d3, 1e-10);
        VectorComposite dc(d, "PVpvLO", {3, 3, 3, 3, 3, 4});
        VectorComposite d2c(d2, "PVpvLO", {3, 3, 3, 3, 3, 4});
        VectorComposite d3c(d3, "PVpvLO", {3, 3, 3, 3, 3, 4});
        VectorXd        de  = dc.vector("pvLO");
        VectorXd        d2e = d2c.vector("pvLO");
        VectorXd        d3e = d3c.vector("pvLO");
        ASSERT_MATRIX_APPROX(d2c.vector("pvLO"), dc.vector("pvLO"), 1e-10);
        ASSERT_MATRIX_APPROX(d3c.vector("pvLO"), dc.vector("pvLO"), 1e-10);
    }

    composeOverState(x, d, dt, x2);
    VectorComposite ddc(d, "PVpvLO", {3, 3, 3, 3, 3, 4});
    ddc.at('p') = ddc.at('P');
    ddc.at('v') = ddc.at('V');
    ASSERT_MATRIX_APPROX(betweenStates(x, x2, dt), ddc.vector("PVpvLO"), 1e-10);

    // x + (x2 - x) = x2
    ASSERT_MATRIX_APPROX(composeOverState(x, betweenStates(x, x2, dt), dt), x2, 1e-10);
}

TEST(FTI_tools, compose_between_with_state_composite)
{
    double dt = 0.1;

    VectorComposite x(VectorXd::Random(13), "PVLO", {3,3,3,4});
    x.at('O').normalize();    
    VectorComposite d(VectorXd::Random(19), "PVpvLO", {3,3,3,3,3,4});
    d.at('O').normalize();    

    VectorComposite xi = composeOverState(x, d, dt, fti::IMU);  
    VectorComposite xf = composeOverState(x, d, dt, fti::FT);  

    ASSERT_MATRIX_APPROX(xi.at('L'), xf.at('L'), 1e-10);
    ASSERT_MATRIX_APPROX(xi.at('O'), xf.at('O'), 1e-10);

    // Obtained delta must conform to IMU "PV"
    VectorComposite d2 = betweenStates(x, xi, dt);

    ASSERT_MATRIX_APPROX(d2.at('P'), d.at('P'), 1e-10);
    ASSERT_MATRIX_APPROX(d2.at('V'), d.at('V'), 1e-10);
    ASSERT_MATRIX_APPROX(d2.at('p'), d.at('P'), 1e-10);
    ASSERT_MATRIX_APPROX(d2.at('v'), d.at('V'), 1e-10);

    // Obtained delta must conform to FT "pv"
    d2 = betweenStates(x, xf, dt);

    ASSERT_MATRIX_APPROX(d2.at('P'), d.at('p'), 1e-10);
    ASSERT_MATRIX_APPROX(d2.at('V'), d.at('v'), 1e-10);
    ASSERT_MATRIX_APPROX(d2.at('p'), d.at('p'), 1e-10);
    ASSERT_MATRIX_APPROX(d2.at('v'), d.at('v'), 1e-10);
}

TEST(FTI_tools, exp_log)
{
    VectorXd tangent(18);
    tangent << VectorXd::Random(15), .1 * Vector3d::Random();  // use angles in the ball theta < pi

    // transform to delta
    VectorXd delta = exp_fti(tangent);

    // expected delta
    Vector3d    dpi = tangent.segment<3>(0);
    Vector3d    dvi = tangent.segment<3>(3);
    Vector3d    dpd = tangent.segment<3>(6);
    Vector3d    dvd = tangent.segment<3>(9);
    Vector3d    dL  = tangent.segment<3>(12);
    Quaterniond dq  = v2q(tangent.segment<3>(15));
    VectorXd    delta_true(19);
    delta_true << dpi, dvi, dpd, dvd, dL, dq.coeffs();
    ASSERT_MATRIX_APPROX(delta, delta_true, 1e-10);

    // transform to a new tangent -- should be the original tangent
    VectorXd d_from_delta = log_fti(delta);
    ASSERT_MATRIX_APPROX(d_from_delta, tangent, 1e-10);

    // transform to a new delta -- should be the previous delta
    VectorXd delta_from_d = exp_fti(d_from_delta);
    ASSERT_MATRIX_APPROX(delta_from_d, delta, 1e-10);
}

TEST(FTI_tools, plus)
{
    VectorXd d1(19), d2(19), d3(19);
    VectorXd err(18);
    Vector4d qv = Vector4d::Random().normalized();
    d1 << VectorXd::Random(15), qv;
    err << Matrix<double, 18, 1>::Random();

    d3.head<3>()      = d1.head<3>() + err.head<3>();
    d3.segment<3>(3)  = d1.segment<3>(3) + err.segment<3>(3);
    d3.segment<3>(6)  = d1.segment<3>(6) + err.segment<3>(6);
    d3.segment<3>(9)  = d1.segment<3>(9) + err.segment<3>(9);
    d3.segment<3>(12) = d1.segment<3>(12) + err.segment<3>(12);
    d3.tail<4>()      = (Quaterniond(qv.data()) * exp_q(err.tail<3>())).coeffs();

    plus(d1, err, d2);
    ASSERT_MATRIX_APPROX(diff(d3, d2), VectorXd::Zero(18), 1e-10);
}

TEST(FTI_tools, diff)
{
    VectorXd d1(19), d2(19);
    VectorXd err(18);
    Vector4d qv = Vector4d::Random().normalized();
    d1 << VectorXd::Random(15), qv;
    d2 = d1;

    diff(d1, d2, err);
    ASSERT_MATRIX_APPROX(err, VectorXd::Zero(18), 1e-10);
    ASSERT_MATRIX_APPROX(diff(d1, d2), VectorXd::Zero(18), 1e-10);

    VectorXd d3(19);
    d3.setRandom();
    d3.tail<4>().normalize();
    err.head<3>()      = d3.head<3>() - d1.head<3>();
    err.segment<3>(3)  = d3.segment<3>(3) - d1.segment<3>(3);
    err.segment<3>(6)  = d3.segment<3>(6) - d1.segment<3>(6);
    err.segment<3>(9)  = d3.segment<3>(9) - d1.segment<3>(9);
    err.segment<3>(12) = d3.segment<3>(12) - d1.segment<3>(12);
    err.tail<3>()      = log_q(Quaterniond(d1.data() + 15).conjugate() * Quaterniond(d3.data() + 15));

    ASSERT_MATRIX_APPROX(err, diff(d1, d3), 1e-10);
}

TEST(FTI_tools, compose_jacobians)
{
    VectorXd d1(19), d2(19), d3(19), d1_pert(19), d2_pert(19), d3_pert(19), d3_pert_approx(19);  // deltas
    VectorXd h(18), j_h(18);  // tangent perturbation
    Matrix<double, 18, 18> J_d3_d1, J_d3_d2, J_d3_d1_num, J_d3_d2_num;
    Vector4d               qv1, qv2;
    double                 dt = 0.1;
    double                 dx = 1e-6;
    qv1                       = Vector4d::Random().normalized();
    d1 << VectorXd::Random(15), qv1;
    qv2 = Vector4d::Random().normalized();
    d2 << VectorXd::Random(15), qv2;

    // analytical jacobians
    compose(d1, d2, dt, d3, J_d3_d1, J_d3_d2);

    ASSERT_MATRIX_APPROX(d3, compose(d1, d2, dt), 1e-12);

    // Check Jacobians using f( x (+) h ) ~ f(x) (+) J_x_d * h
    // this implies
    //  - compose( d1(+)h , d2 ) ~ compose( d1, d2) (+) J_d3_d1*h
    //  - compose( d1 , d2(+)h ) ~ compose( d1, d2) (+) J_d3_d2*h

    h = dx * VectorXd::Random(18);

    // Jac wrt. d1
    d1_pert        = plus(d1, h);
    d3_pert        = compose(d1_pert, d2, dt);
    j_h            = J_d3_d1 * h;
    d3_pert_approx = plus(d3, j_h);

    ASSERT_MATRIX_APPROX(d3_pert, d3_pert_approx, 1e-8);

    // Jac wrt. d2
    d2_pert        = plus(d2, h);
    d3_pert        = compose(d1, d2_pert, dt);
    j_h            = J_d3_d2 * h;
    d3_pert_approx = plus(d3, j_h);

    ASSERT_MATRIX_APPROX(d3_pert, d3_pert_approx, 1e-12);
}

TEST(FTI_tools, data2tangent_hover)
{
    Vector3d am(1, 2, 3), wm(1, 2, 3), ab(0, 0, 0), wb(0, 0, 0), fb(0, 0, 0), mb(0, 0, 0), r_bc(0, 0, 0);

    std::list<Propeller> propellers;
    propellers.emplace_back(Vector3d(1, 1, 0), Matrix3d::Identity(), 1, 1, 1);    // FL
    propellers.emplace_back(Vector3d(-1, 1, 0), Matrix3d::Identity(), -1, 1, 1);  // RL
    propellers.emplace_back(Vector3d(-1, -1, 0), Matrix3d::Identity(), 1, 1, 1);  // RR
    propellers.emplace_back(Vector3d(1, -1, 0), Matrix3d::Identity(), -1, 1, 1);  // FR

    Vector3d a, w, f, m;

    Vector4d n(2, 2, 2, 2);  // FL, RL, RR, FR

    data2tangent(am, wm, n, ab, wb, fb, mb, r_bc, propellers, a, w, f, m);

    ASSERT_MATRIX_APPROX(a, am - ab, 1e-12);
    ASSERT_MATRIX_APPROX(w, wm - wb, 1e-12);
    ASSERT_MATRIX_APPROX(f, Vector3d(0, 0, 16), 1e-12);
    ASSERT_MATRIX_APPROX(m, Vector3d::Zero(), 1e-12);
}

TEST(FTI_tools, data2tangent_yawing)
{
    Vector3d am(1, 2, 3), wm(1, 2, 3), ab(0, 0, 0), wb(0, 0, 0), fb(0, 0, 0), mb(0, 0, 0), r_bc(0, 0, 0);

    std::list<Propeller> propellers;
    propellers.emplace_back(Vector3d(1, 1, 0), Matrix3d::Identity(), 1, 1, 1);    // FL
    propellers.emplace_back(Vector3d(-1, 1, 0), Matrix3d::Identity(), -1, 1, 1);  // RL
    propellers.emplace_back(Vector3d(-1, -1, 0), Matrix3d::Identity(), 1, 1, 1);  // RR
    propellers.emplace_back(Vector3d(1, -1, 0), Matrix3d::Identity(), -1, 1, 1);  // FR

    Vector3d a, w, f, m;

    Vector4d n(2, 0, 2, 0);  // FL, RL, RR, FR

    data2tangent(am, wm, n, ab, wb, fb, mb, r_bc, propellers, a, w, f, m);

    ASSERT_MATRIX_APPROX(a, am - ab, 1e-12);
    ASSERT_MATRIX_APPROX(w, wm - wb, 1e-12);
    ASSERT_MATRIX_APPROX(f, Vector3d(0, 0, 8), 1e-12);
    ASSERT_MATRIX_APPROX(m, Vector3d(0, 0, -8), 1e-12);
}

TEST(FTI_tools, data2tangent_jacobians)
{
    VectorXd data(10), bias(12), tangent(12);
    Vector3d r_bc;

    std::list<Propeller> propellers;
    propellers.emplace_back(Vector3d(1, 1, 0), Matrix3d::Identity(), 1, 1, 1);    // FL
    propellers.emplace_back(Vector3d(-1, 1, 0), Matrix3d::Identity(), -1, 1, 1);  // RL
    propellers.emplace_back(Vector3d(-1, -1, 0), Matrix3d::Identity(), 1, 1, 1);  // RR
    propellers.emplace_back(Vector3d(1, -1, 0), Matrix3d::Identity(), -1, 1, 1);  // FR

    MatrixXd J_tan_data(12, 10), J_tan_bias(12, 12), J_tan_com(12, 3);

    data.setRandom();
    bias.setRandom();
    r_bc.setRandom();

    data2tangent(data, bias, r_bc, propellers, tangent, J_tan_data, J_tan_bias, J_tan_com);

    // all jacobians
    VectorXd tangent_pert(12), tangent_pert_aprox(12);
    double   dx = 1e-6;
    VectorXd pert;

    // jac wrt data
    pert               = data.Random(10) * dx;
    VectorXd data_pert = data + pert;
    data2tangent(data_pert, bias, r_bc, propellers, tangent_pert);
    tangent_pert_aprox = tangent + J_tan_data * pert;

    ASSERT_MATRIX_APPROX(tangent_pert, tangent_pert_aprox, 1e-10);

    // jac wrt bias
    pert               = bias.Random(12) * dx;
    VectorXd bias_pert = bias + pert;
    data2tangent(data, bias_pert, r_bc, propellers, tangent_pert);
    tangent_pert_aprox = tangent + J_tan_bias * pert;

    ASSERT_MATRIX_APPROX(tangent_pert, tangent_pert_aprox, 1e-10);

    // jac wrt com
    pert               = r_bc.Random(3) * dx;
    VectorXd r_bc_pert = r_bc + pert;
    data2tangent(data, bias, r_bc_pert, propellers, tangent_pert);
    tangent_pert_aprox = tangent + J_tan_com * pert;

    ASSERT_MATRIX_APPROX(tangent_pert, tangent_pert_aprox, 1e-10);
}

TEST(FTI_tools, forces2acc_onlyforce)
{
    Vector3d force(1, 2, 3);
    Vector3d torque(0, 0, 0);
    Vector3d angvel(0, 0, 0);
    Vector3d com(0, 0, 0);
    Vector3d inertia(1, 1, 1);
    double   mass(10);
    Vector3d acc;

    //
    forces2acc(force, torque, angvel, com, inertia, mass, acc);
    ASSERT_MATRIX_APPROX(acc, force / mass, 1e-12);

    //
    angvel << 3, 2, 1;

    forces2acc(force, torque, angvel, com, inertia, mass, acc);
    ASSERT_MATRIX_APPROX(acc, force / mass, 1e-12);

    //
    angvel << 0, 0, 0;
    com << 1, 2, 3;

    forces2acc(force, torque, angvel, com, inertia, mass, acc);
    ASSERT_MATRIX_APPROX(acc, force / mass, 1e-12);

    // must not pass
    angvel << 3, 2, 1;
    com << 1, 2, 3;

    forces2acc(force, torque, angvel, com, inertia, mass, acc);
    //    ASSERT_MATRIX_APPROX(acc, force/mass, 1e-12);
}

TEST(FTI_tools, forces2acc_jacobians_matlab)
{
    Vector3d force(1, 2, 3);
    Vector3d torque(4, 5, 6);
    Vector3d angvel(3, 2, 1);
    Vector3d com(6, 5, 4);
    Vector3d inertia(1, -2, 3);
    double   mass = 2;

    Vector3d acc;
    Matrix3d J_acc_force, J_acc_torque, J_acc_angvel, J_acc_com, J_acc_inertia; 
    Vector3d J_acc_mass;

    forces2acc(force, torque, angvel, com, inertia, mass, 
                acc, 
                J_acc_force, J_acc_torque, J_acc_angvel, J_acc_com,
                J_acc_inertia, J_acc_mass);

    // Matlab results using symbolic Jacobians:
    //
    // acc =
    //    50.5000
    //   -65.0000
    //    22.5000
    // J_acc_force =
    //     0.5000         0         0
    //          0    0.5000         0
    //          0         0    0.5000
    // J_acc_torque =
    //          0    2.0000    1.6667
    //     4.0000         0   -2.0000
    //    -5.0000   -3.0000         0
    // J_acc_angvel =
    //      0    24    12
    //      6   -60   -38
    //     12    36     4
    // J_acc_com =
    //     5.0000    2.0000    2.5000
    //   -14.0000   10.0000   -8.0000
    //    -8.5000    4.0000   13.0000
    // J_acc_inertia =
    //     4.0000    1.0000   -7.3333
    //    12.0000   20.0000    8.0000
    //   -21.0000  -26.5000    1.0000
    // J_acc_mass =
    //    -0.2500
    //    -0.5000
    //    -0.7500

    WOLF_INFO("acc: \n", acc);
    WOLF_INFO("J_acc_force: \n", J_acc_force);
    WOLF_INFO("J_acc_torque: \n", J_acc_torque);
    WOLF_INFO("J_acc_angvel: \n", J_acc_angvel);
    WOLF_INFO("J_acc_com: \n", J_acc_com);
    WOLF_INFO("J_acc_inertia: \n", J_acc_inertia);
    WOLF_INFO("J_acc_mass: \n", J_acc_mass);
}

TEST(FTI_tools, forces2acc_jacobians)
{
    Vector3d force, torque, angvel, com, inertia;
    double   mass;
    Vector3d acc;

    force   .setRandom().normalize();
    torque  .setRandom().normalize();
    angvel  .setRandom().normalize();
    com     .setRandom().normalize();
    inertia .setRandom().normalize(); inertia *= 100;

    inertia = (inertia.array() * inertia.array()).matrix();  // make positive
    mass    = 2.0;

    MatrixXd J_acc_force, J_acc_torque, J_acc_angvel, J_acc_com, J_acc_inertia, J_acc_mass;

    // analytical jacs
    forces2acc(force, torque, angvel, com, inertia, mass, acc, J_acc_force, J_acc_torque, J_acc_angvel, J_acc_com,
               J_acc_inertia, J_acc_mass);

    // linear approximations
    Vector3d pert, acc_pert, acc_pert_approx;
    double   dx = 1e-5;
    pert        = dx * Vector3d::Random().normalized();

    // force
    forces2acc(force + pert, torque, angvel, com, inertia, mass, acc_pert);
    acc_pert_approx = acc + J_acc_force * pert;

    ASSERT_MATRIX_APPROX(acc_pert, acc_pert_approx, 1e-8);

    // torque
    forces2acc(force, torque + pert, angvel, com, inertia, mass, acc_pert);
    acc_pert_approx = acc + J_acc_torque * pert;

    ASSERT_MATRIX_APPROX(acc_pert, acc_pert_approx, 1e-8);

    // angvel
    forces2acc(force, torque, angvel + pert, com, inertia, mass, acc_pert);
    acc_pert_approx = acc + J_acc_angvel * pert;

    ASSERT_MATRIX_APPROX(acc_pert, acc_pert_approx, 1e-8);

    // com
    forces2acc(force, torque, angvel, com + pert, inertia, mass, acc_pert);
    acc_pert_approx = acc + J_acc_com * pert;

    ASSERT_MATRIX_APPROX(acc_pert, acc_pert_approx, 1e-8);

    // inertia
    forces2acc(force, torque, angvel, com, inertia + pert, mass, acc_pert);
    acc_pert_approx = acc + J_acc_inertia * pert;

    ASSERT_MATRIX_APPROX(acc_pert, acc_pert_approx, 1e-8);

    // mass
    forces2acc(force, torque, angvel, com, inertia, mass + dx, acc_pert);
    acc_pert_approx = acc + J_acc_mass * dx;

    ASSERT_MATRIX_APPROX(acc_pert, acc_pert_approx, 1e-8);
}

TEST(FTI_tools, tangent2delta_jacobians)
{
    VectorXd tangent(12), delta(19), model(7);
    MatrixXd J_delta_tangent(18, 12), J_delta_model(18, 7);
    double   dt = 1;

    tangent.setRandom();
    model.setRandom();
    model.tail(4) *= 10; // augment inertia and mass

    tangent2delta(tangent, model, dt, delta, J_delta_tangent, J_delta_model);

    // linear approximations
    VectorXd pert, delta_pert(19), delta_pert_approx(19), tangent_pert(12), model_pert(7);
    double   dx = 1e-6;

    // tangent
    pert         = dx * tangent.Random(12);
    tangent_pert = tangent + pert;
    tangent2delta(tangent_pert, model, dt, delta_pert);
    VectorXd step     = J_delta_tangent * pert;
    delta_pert_approx = plus(delta, step);

    ASSERT_MATRIX_APPROX(delta_pert, delta_pert_approx, 1e-9);

    // model
    pert       = dx * model.Random(7);
    model_pert = model + pert;
    tangent2delta(tangent, model_pert, dt, delta_pert);
    step              = J_delta_model * pert;
    delta_pert_approx = plus(delta, step);

    ASSERT_MATRIX_APPROX(delta_pert, delta_pert_approx, 1e-9);
}

TEST(FTI_tools, data2delta_hovering)
{
    VectorXd data(10), bias(12), model(7), delta(19);
    double   dt = 1;

    std::list<Propeller> propellers;
    propellers.emplace_back(Vector3d(1, 1, 0), Matrix3d::Identity(), 1, 1, 1);    // FL
    propellers.emplace_back(Vector3d(-1, 1, 0), Matrix3d::Identity(), -1, 1, 1);  // RL
    propellers.emplace_back(Vector3d(-1, -1, 0), Matrix3d::Identity(), 1, 1, 1);  // RR
    propellers.emplace_back(Vector3d(1, -1, 0), Matrix3d::Identity(), -1, 1, 1);  // FR

    Vector4d n(2, 2, 2, 2);  // FL, RL, RR, FR

    data << 0, 0, 0, 0, 0, 0, n;  // acc, gyro, propeller speeds
    bias.setZero();
    model << 0, 0, 0, 1, 1, 1, 1;  // com, inertia, mass

    data2delta(data, bias, model, propellers, dt, delta);

    VectorXd delta_exp(19);
    delta_exp << 0, 0, 0, 0, 0, 0, 0, 0, 4 * 0.5 * 2 * 2, 0, 0, 4 * 2 * 2, 0, 0, 0, 0, 0, 0, 1;

    ASSERT_MATRIX_APPROX(delta, delta_exp, 1e-12);
}

TEST(FTI_tools, data2delta_jacobians)
{
    VectorXd data(10), bias(12), model(7), delta(19);
    double   dt = 1;
    MatrixXd J_delta_data(18, 10), J_delta_bias(18, 12), J_delta_model(18, 7);

    std::list<Propeller> propellers;
    propellers.emplace_back(Vector3d(1, 1, 0), Matrix3d::Identity(), 1, 1, 1);    // FL
    propellers.emplace_back(Vector3d(-1, 1, 0), Matrix3d::Identity(), -1, 1, 1);  // RL
    propellers.emplace_back(Vector3d(-1, -1, 0), Matrix3d::Identity(), 1, 1, 1);  // RR
    propellers.emplace_back(Vector3d(1, -1, 0), Matrix3d::Identity(), -1, 1, 1);  // FR

    VectorXd pert, step, data_pert(10), bias_pert(12), model_pert(7), delta_pert(19), delta_pert_approx(19);
    double   dx = 1e-6;

    data.setRandom();  // acc, gyro, propeller speeds
    bias.setRandom();
    model.setRandom();  // com, inertia, mass

    data2delta(data, bias, model, propellers, dt, delta, J_delta_data, J_delta_bias, J_delta_model);

    // data
    pert      = dx * data.Random(10);
    data_pert = data + pert;
    data2delta(data_pert, bias, model, propellers, dt, delta_pert);
    step              = J_delta_data * pert;
    delta_pert_approx = plus(delta, step);

    ASSERT_MATRIX_APPROX(delta_pert, delta_pert_approx, 1e-8);

    // bias
    pert      = dx * bias.Random(12);
    bias_pert = bias + pert;
    data2delta(data, bias_pert, model, propellers, dt, delta_pert);
    step              = J_delta_bias * pert;
    delta_pert_approx = plus(delta, step);

    ASSERT_MATRIX_APPROX(delta_pert, delta_pert_approx, 1e-8);

    // model
    pert       = dx * model.Random(7);
    model_pert = model + pert;
    data2delta(data, bias, model_pert, propellers, dt, delta_pert);
    step              = J_delta_model * pert;
    delta_pert_approx = plus(delta, step);

    ASSERT_MATRIX_APPROX(delta_pert, delta_pert_approx, 1e-8);
}


int main(int argc, char **argv)
{
    testing::InitGoogleTest(&argc, argv);
    //  ::testing::GTEST_FLAG(filter) = "FTI_tools.identity*";
    return RUN_ALL_TESTS();
}
